{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# SepViT: Separable Vison Transformer Code"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Helpers"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from functools import partial\n",
    "\n",
    "import torch\n",
    "from torch import nn, einsum\n",
    "\n",
    "from einops import rearrange, repeat\n",
    "from einops.layers.torch import Rearrange, Reduce\n",
    "\n",
    "# helpers\n",
    "\n",
    "def cast_tuple(val, length = 1):\n",
    "    return val if isinstance(val, tuple) else ((val,) * length)\n",
    "\n",
    "# helper classes\n",
    "\n",
    "class ChanLayerNorm(nn.Module):\n",
    "    def __init__(self, dim, eps = 1e-5):\n",
    "        super().__init__()\n",
    "        self.eps = eps\n",
    "        self.g = nn.Parameter(torch.ones(1, dim, 1, 1))\n",
    "        self.b = nn.Parameter(torch.zeros(1, dim, 1, 1))\n",
    "\n",
    "    def forward(self, x):\n",
    "        var = torch.var(x, dim = 1, unbiased = False, keepdim = True)\n",
    "        mean = torch.mean(x, dim = 1, keepdim = True)\n",
    "        return (x - mean) / (var + self.eps).sqrt() * self.g + self.b\n",
    "\n",
    "class PreNorm(nn.Module):\n",
    "    def __init__(self, dim, fn):\n",
    "        super().__init__()\n",
    "        self.norm = ChanLayerNorm(dim)\n",
    "        self.fn = fn\n",
    "\n",
    "    def forward(self, x):\n",
    "        return self.fn(self.norm(x))\n",
    "\n",
    "class OverlappingPatchEmbed(nn.Module):\n",
    "    def __init__(self, dim_in, dim_out, stride = 2):\n",
    "        super().__init__()\n",
    "        kernel_size = stride * 2 - 1\n",
    "        padding = kernel_size // 2\n",
    "        self.conv = nn.Conv2d(dim_in, dim_out, kernel_size, stride = stride, padding = padding)\n",
    "\n",
    "    def forward(self, x):\n",
    "        return self.conv(x)\n",
    "\n",
    "class PEG(nn.Module):\n",
    "    def __init__(self, dim, kernel_size = 3):\n",
    "        super().__init__()\n",
    "        self.proj = nn.Conv2d(dim, dim, kernel_size = kernel_size, padding = kernel_size // 2, groups = dim, stride = 1)\n",
    "\n",
    "    def forward(self, x):\n",
    "        return self.proj(x) + x"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Transformer Module"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# feedforward\n",
    "\n",
    "class FeedForward(nn.Module):\n",
    "    def __init__(self, dim, mult = 4, dropout = 0.):\n",
    "        super().__init__()\n",
    "        inner_dim = int(dim * mult)\n",
    "        self.net = nn.Sequential(\n",
    "            nn.Conv2d(dim, inner_dim, 1),\n",
    "            nn.GELU(),\n",
    "            nn.Dropout(dropout),\n",
    "            nn.Conv2d(inner_dim, dim, 1),\n",
    "            nn.Dropout(dropout)\n",
    "        )\n",
    "    def forward(self, x):\n",
    "        return self.net(x)\n",
    "\n",
    "# attention\n",
    "\n",
    "class DSSA(nn.Module):\n",
    "    def __init__(\n",
    "        self,\n",
    "        dim,\n",
    "        heads = 8,\n",
    "        dim_head = 32,\n",
    "        dropout = 0.,\n",
    "        window_size = 7\n",
    "    ):\n",
    "        super().__init__()\n",
    "        self.heads = heads\n",
    "        self.scale = dim_head ** -0.5\n",
    "        self.window_size = window_size\n",
    "        inner_dim = dim_head * heads\n",
    "\n",
    "        self.attend = nn.Sequential(\n",
    "            nn.Softmax(dim = -1),\n",
    "            nn.Dropout(dropout)\n",
    "        )\n",
    "\n",
    "        self.to_qkv = nn.Conv1d(dim, inner_dim * 3, 1, bias = False)\n",
    "\n",
    "        # window tokens\n",
    "\n",
    "        self.window_tokens = nn.Parameter(torch.randn(dim))\n",
    "\n",
    "        # prenorm and non-linearity for window tokens\n",
    "        # then projection to queries and keys for window tokens\n",
    "\n",
    "        self.window_tokens_to_qk = nn.Sequential(\n",
    "            nn.LayerNorm(dim_head),\n",
    "            nn.GELU(),\n",
    "            Rearrange('b h n c -> b (h c) n'),\n",
    "            nn.Conv1d(inner_dim, inner_dim * 2, 1),\n",
    "            Rearrange('b (h c) n -> b h n c', h = heads),\n",
    "        )\n",
    "\n",
    "        # window attention\n",
    "\n",
    "        self.window_attend = nn.Sequential(\n",
    "            nn.Softmax(dim = -1),\n",
    "            nn.Dropout(dropout)\n",
    "        )\n",
    "\n",
    "        self.to_out = nn.Sequential(\n",
    "            nn.Conv2d(inner_dim, dim, 1),\n",
    "            nn.Dropout(dropout)\n",
    "        )\n",
    "\n",
    "    def forward(self, x):\n",
    "        \"\"\"\n",
    "        einstein notation\n",
    "\n",
    "        b - batch\n",
    "        c - channels\n",
    "        w1 - window size (height)\n",
    "        w2 - also window size (width)\n",
    "        i - sequence dimension (source)\n",
    "        j - sequence dimension (target dimension to be reduced)\n",
    "        h - heads\n",
    "        x - height of feature map divided by window size\n",
    "        y - width of feature map divided by window size\n",
    "        \"\"\"\n",
    "\n",
    "        batch, height, width, heads, wsz = x.shape[0], *x.shape[-2:], self.heads, self.window_size\n",
    "        assert (height % wsz) == 0 and (width % wsz) == 0, f'height {height} and width {width} must be divisible by window size {wsz}'\n",
    "        num_windows = (height // wsz) * (width // wsz)\n",
    "        # x.shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "\n",
    "        # fold in windows for \"depthwise\" attention - not sure why it is named depthwise when it is just \"windowed\" attention\n",
    "\n",
    "        x = rearrange(x, 'b c (h w1) (w w2) -> (b h w) c (w1 w2)', w1 = wsz, w2 = wsz)\n",
    "        # x.shape : torch.Size([64, 32, 49]) / torch.Size([16, 64, 49]) / torch.Size([4, 128, 49]) / torch.Size([1, 256, 49])\n",
    "\n",
    "        # add windowing tokens\n",
    "\n",
    "        w = repeat(self.window_tokens, 'c -> b c 1', b = x.shape[0])\n",
    "        # w.shape : torch.Size([64, 32, 1]) / torch.Size([16, 64, 1]) / torch.Size([4, 128, 1]) / torch.Size([1, 256, 1])\n",
    "        x = torch.cat((w, x), dim = -1)\n",
    "        # x.shape : torch.Size([64, 32, 50]) / torch.Size([16, 64, 50]) / torch.Size([4, 128, 50]) / torch.Size([1, 256, 50])\n",
    "\n",
    "        # project for queries, keys, value\n",
    "\n",
    "        q, k, v = self.to_qkv(x).chunk(3, dim = 1)\n",
    "        # q.shape : torch.Size([64, 32, 50]) / torch.Size([16, 64, 50]) / torch.Size([4, 128, 50]) / torch.Size([1, 256, 50])\n",
    "\n",
    "        # split out heads\n",
    "\n",
    "        q, k, v = map(lambda t: rearrange(t, 'b (h d) ... -> b h (...) d', h = heads), (q, k, v))\n",
    "        # q.shape : torch.Size([64, 1, 50, 32]) / torch.Size([16, 2, 50, 32]) / torch.Size([4, 4, 50, 32]) / torch.Size([1, 8, 50, 32])\n",
    "\n",
    "        # scale\n",
    "\n",
    "        q = q * self.scale\n",
    "\n",
    "        # similarity\n",
    "\n",
    "        dots = einsum('b h i d, b h j d -> b h i j', q, k)\n",
    "        # dots.shape : torch.Size([64, 1, 50, 50]) / torch.Size([16, 2, 50, 50]) / torch.Size([4, 4, 50, 50]) / torch.Size([1, 8, 50, 50])\n",
    "\n",
    "        # attention\n",
    "\n",
    "        attn = self.attend(dots)\n",
    "        # attn.shape : torch.Size([64, 1, 50, 50]) / torch.Size([16, 2, 50, 50]) / torch.Size([4, 4, 50, 50]) / torch.Size([1, 8, 50, 50])\n",
    "        \n",
    "        # aggregate values\n",
    "\n",
    "        out = torch.matmul(attn, v)\n",
    "        # out.shape : torch.Size([64, 1, 50, 32]) / torch.Size([16, 2, 50, 32]) / torch.Size([4, 4, 50, 32]) / torch.Size([1, 8, 50, 32])\n",
    "\n",
    "        # split out windowed tokens\n",
    "\n",
    "        window_tokens, windowed_fmaps = out[:, :, 0], out[:, :, 1:]\n",
    "        # window_tokens.shape : torch.Size([64, 1, 32]) / torch.Size([16, 2, 32]) / torch.Size([4, 4, 32]) / torch.Size([1, 8, 32])\n",
    "        # windowed_fmaps.shape : torch.Size([64, 1, 49, 32]) / torch.Size([16, 2, 49, 32]) / torch.Size([4, 4, 49, 32]) / torch.Size([1, 8, 49, 32])\n",
    "\n",
    "        # early return if there is only 1 window\n",
    "\n",
    "        if num_windows == 1:\n",
    "            fmap = rearrange(windowed_fmaps, '(b x y) h (w1 w2) d -> b (h d) (x w1) (y w2)', x = height // wsz, y = width // wsz, w1 = wsz, w2 = wsz)\n",
    "            return self.to_out(fmap)\n",
    "\n",
    "        # carry out the pointwise attention, the main novelty in the paper\n",
    "\n",
    "        window_tokens = rearrange(window_tokens, '(b x y) h d -> b h (x y) d', x = height // wsz, y = width // wsz)\n",
    "        # window_tokens.shape : torch.Size([1, 1, 64, 32]) / torch.Size([1, 2, 16, 32]) / torch.Size([1, 4, 4, 32]) / torch.Size([1, 8, 1, 32])\n",
    "        windowed_fmaps = rearrange(windowed_fmaps, '(b x y) h n d -> b h (x y) n d', x = height // wsz, y = width // wsz)\n",
    "        # windowed_fmaps.shape : torch.Size([1, 1, 64, 49, 32]) / torch.Size([1, 2, 16, 49, 32]) / torch.Size([1, 4, 4, 49, 32]) /  / torch.Size([1, 8, 1, 32])\n",
    "\n",
    "        # windowed queries and keys (preceded by prenorm activation)\n",
    "\n",
    "        w_q, w_k = self.window_tokens_to_qk(window_tokens).chunk(2, dim = -1)\n",
    "        # w_q.shape : torch.Size([1, 1, 64, 32]) / torch.Size([1, 2, 16, 32]) / torch.Size([1, 4, 4, 32]) / torch.Size([1, 8, 1, 32])\n",
    "        # w_k.shape : torch.Size([1, 1, 64, 32]) / torch.Size([1, 2, 16, 32]) / torch.Size([1, 4, 4, 32]) / torch.Size([1, 8, 1, 32])\n",
    "\n",
    "        # scale\n",
    "\n",
    "        w_q = w_q * self.scale\n",
    "        # similarities\n",
    "\n",
    "        w_dots = einsum('b h i d, b h j d -> b h i j', w_q, w_k)\n",
    "        # w_dots : torch.Size([1, 1, 64, 64]) / torch.Size([1, 2, 16, 16]) / torch.Size([1, 4, 4, 4]) / torch.Size([1, 8, 1, 1])\n",
    "        w_attn = self.window_attend(w_dots)\n",
    "        # w_attn : torch.Size([1, 1, 64, 64]) / torch.Size([1, 2, 16, 16]) / torch.Size([1, 4, 4, 4]) / torch.Size([1, 8, 1, 1])\n",
    "\n",
    "        # aggregate the feature maps from the \"depthwise\" attention step (the most interesting part of the paper, one i haven't seen before)\n",
    "\n",
    "        aggregated_windowed_fmap = einsum('b h i j, b h j w d -> b h i w d', w_attn, windowed_fmaps)\n",
    "        # aggregated_windowed_fmap.shape : torch.Size([1, 1, 64, 49, 32]) / torch.Size([1, 2, 16, 49, 32]) / torch.Size([1, 4, 4, 49, 32]) / torch.Size([1, 8, 1, 49, 32])\n",
    "\n",
    "        # fold back the windows and then combine heads for aggregation\n",
    "\n",
    "        fmap = rearrange(aggregated_windowed_fmap, 'b h (x y) (w1 w2) d -> b (h d) (x w1) (y w2)', x = height // wsz, y = width // wsz, w1 = wsz, w2 = wsz)\n",
    "        # fmap.shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "        return self.to_out(fmap)\n",
    "        # self.to_out(fmap).shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "\n",
    "class Transformer(nn.Module):\n",
    "    def __init__(\n",
    "        self,\n",
    "        dim,\n",
    "        depth,\n",
    "        dim_head = 32,\n",
    "        heads = 8,\n",
    "        ff_mult = 4,\n",
    "        dropout = 0.,\n",
    "        norm_output = True\n",
    "    ):\n",
    "        super().__init__()\n",
    "        self.layers = nn.ModuleList([])\n",
    "\n",
    "        for ind in range(depth):\n",
    "            self.layers.append(nn.ModuleList([\n",
    "                PreNorm(dim, DSSA(dim, heads = heads, dim_head = dim_head, dropout = dropout)),\n",
    "                PreNorm(dim, FeedForward(dim, mult = ff_mult, dropout = dropout)),\n",
    "            ]))\n",
    "\n",
    "        self.norm = ChanLayerNorm(dim) if norm_output else nn.Identity()\n",
    "\n",
    "    def forward(self, x):\n",
    "        for attn, ff in self.layers:\n",
    "            x = attn(x) + x\n",
    "            x = ff(x) + x\n",
    "\n",
    "        return self.norm(x)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## SepViT"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SepViT(nn.Module):\n",
    "    def __init__(\n",
    "        self,\n",
    "        *,\n",
    "        num_classes,\n",
    "        dim,\n",
    "        depth,\n",
    "        heads,\n",
    "        window_size = 7,\n",
    "        dim_head = 32,\n",
    "        ff_mult = 4,\n",
    "        channels = 3,\n",
    "        dropout = 0.\n",
    "    ):\n",
    "        super().__init__()\n",
    "        assert isinstance(depth, tuple), 'depth needs to be tuple if integers indicating number of transformer blocks at that stage'\n",
    "\n",
    "        num_stages = len(depth)\n",
    "\n",
    "        dims = tuple(map(lambda i: (2 ** i) * dim, range(num_stages)))\n",
    "        dims = (channels, *dims)\n",
    "        dim_pairs = tuple(zip(dims[:-1], dims[1:]))\n",
    "\n",
    "        strides = (4, *((2,) * (num_stages - 1)))\n",
    "\n",
    "        hyperparams_per_stage = [heads, window_size]\n",
    "        hyperparams_per_stage = list(map(partial(cast_tuple, length = num_stages), hyperparams_per_stage))\n",
    "        assert all(tuple(map(lambda arr: len(arr) == num_stages, hyperparams_per_stage)))\n",
    "\n",
    "        self.layers = nn.ModuleList([])\n",
    "\n",
    "        for ind, ((layer_dim_in, layer_dim), layer_depth, layer_stride, layer_heads, layer_window_size) in enumerate(zip(dim_pairs, depth, strides, *hyperparams_per_stage)):\n",
    "            is_last = ind == (num_stages - 1)\n",
    "\n",
    "            self.layers.append(nn.ModuleList([\n",
    "                OverlappingPatchEmbed(layer_dim_in, layer_dim, stride = layer_stride),\n",
    "                PEG(layer_dim),\n",
    "                Transformer(dim = layer_dim, depth = layer_depth, heads = layer_heads, ff_mult = ff_mult, dropout = dropout, norm_output = not is_last),\n",
    "            ]))\n",
    "\n",
    "        self.mlp_head = nn.Sequential(\n",
    "            Reduce('b d h w -> b d', 'mean'),\n",
    "            nn.LayerNorm(dims[-1]),\n",
    "            nn.Linear(dims[-1], num_classes)\n",
    "        )\n",
    "\n",
    "    def forward(self, x):\n",
    "        # x.shape : torch.Size([1, 3, 224, 224])\n",
    "        for ope, peg, transformer in self.layers: \n",
    "            x = ope(x) # x.shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "            x = peg(x) # x.shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "            x = transformer(x) # x.shape : torch.Size([1, 32, 56, 56]) / torch.Size([1, 64, 28, 28]) / torch.Size([1, 128, 14, 14]) / torch.Size([1, 256, 7, 7])\n",
    "\n",
    "        return self.mlp_head(x)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Main"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "v = SepViT(\n",
    "    num_classes = 1000,\n",
    "    dim = 32,               # dimensions of first stage, which doubles every stage (32, 64, 128, 256) for SepViT-Lite\n",
    "    dim_head = 32,          # attention head dimension\n",
    "    heads = (1, 2, 4, 8),   # number of heads per stage\n",
    "    depth = (1, 2, 6, 2),   # number of transformer blocks per stage\n",
    "    window_size = 7,        # window size of DSS Attention block\n",
    "    dropout = 0.1           # dropout\n",
    ")\n",
    "\n",
    "img = torch.randn(1, 3, 224, 224)\n",
    "\n",
    "preds = v(img) # (1, 1000)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "---\n",
    "\n",
    "Author by `정영상`  \n",
    "Edit by `김주영`"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3.10.4 64-bit",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "name": "python",
   "version": "3.10.4 (v3.10.4:9d38120e33, Mar 23 2022, 17:29:05) [Clang 13.0.0 (clang-1300.0.29.30)]"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "aee8b7b246df8f9039afb4144a1f6fd8d2ca17a180786b69acc140d282b71a49"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
